/*
 *	HT Editor
 *	htstream.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htdebug.h"
#include "htstream.h"
#include "tools.h"

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/stat.h>

/*
 *	CLASS ht_stream
 */

void ht_stream::init()
{
	error=0;
	stream_error_proc=0;
	access_mode=0;
	mode=0;
}

void ht_stream::done()
{
}

int ht_stream::call_error_proc()
{
	if (stream_error_proc) return stream_error_proc(this); else return SERR_FAIL;
}

void ht_stream::flush()
{
}

int ht_stream::get_error()
{
	return error;
}

dword ht_stream::get_mode()
{
	return mode;
}

char *ht_stream::get_name()
{
	return 0;
}

dword ht_stream::get_size()
{
	return 0;
}

void ht_stream::invalidate()
{
}

int ht_stream::isdirty(STREAMOFS offset, int range)
{
	return 0;
}

void ht_stream::pstat(ht_pstat *s)
{
	s->caps=0;
}

dword ht_stream::read(void *buf, dword size)
{
	abstract();
	return 0;
}

int ht_stream::seek(STREAMOFS offset)
{
	return 0;
}

int ht_stream::set_access_mode(dword _access_mode)
{
	access_mode=_access_mode;
	return 1;
}

void ht_stream::set_error(int e)
{
	error=e;
}

void ht_stream::set_error_proc(int (*_stream_error_proc)(ht_stream *stream))
{
	stream_error_proc=_stream_error_proc;
}

void ht_stream::set_mode(dword _mode)
{
	mode=_mode;
}

STREAMOFS ht_stream::tell()
{
	return 0;
}

dword ht_stream::write(void *buf, dword size)
{
	abstract();
	return 0;
}

/*
 *	ht_layer_stream
 */

void ht_layer_stream::init(ht_stream *_stream, int _own_stream)
{
	ht_stream::init();
	stream=_stream;
	own_stream=_own_stream;
}

void ht_layer_stream::done()
{
	if (own_stream) {
		stream->done();
		delete stream;
	}
	ht_stream::done();
}

void ht_layer_stream::flush()
{
	stream->flush();
}

int ht_layer_stream::get_error()
{
	return stream->get_error();
}

dword ht_layer_stream::get_mode()
{
	return mode;
}

char *ht_layer_stream::get_name()
{
	return stream->get_name();
}

dword ht_layer_stream::get_size()
{
	return stream->get_size();
}

void ht_layer_stream::invalidate()
{
	stream->invalidate();
}

int ht_layer_stream::isdirty(STREAMOFS offset, int range)
{
	return stream->isdirty(offset, range);
}

void ht_layer_stream::pstat(ht_pstat *s)
{
	return stream->pstat(s);
}

int ht_layer_stream::seek(STREAMOFS offset)
{
	return stream->seek(offset);
}

int ht_layer_stream::set_access_mode(dword access_mode)
{
	return stream->set_access_mode(access_mode);
}

void ht_layer_stream::set_error_proc(int (*stream_error_proc)(ht_stream *stream))
{
	stream->set_error_proc(stream_error_proc);
}

void ht_layer_stream::set_mode(dword _mode)
{
	mode=_mode;
}

STREAMOFS ht_layer_stream::tell()
{
	return stream->tell();
}

dword ht_layer_stream::read(void *buf, dword size)
{
	return stream->read(buf, size);
}

dword ht_layer_stream::write(void *buf, dword size)
{
	return stream->write(buf, size);
}

/*
 *	CLASS ht_file_stream
 */

void ht_file_stream::init(const char *_name, int _access_mode)
{
	ht_stream::init();
	name=(char*)malloc(strlen(_name)+1);
	strcpy(name, _name);
	offset=0;
	file=0;
	access_mode=0;
	set_access_mode(_access_mode);
}

void ht_file_stream::done()
{
	if (name) delete name;
	if (file) fclose(file);
	ht_stream::done();
}

char *ht_file_stream::get_name()
{
	return name;
}

dword ht_file_stream::get_size()
{
/* FIXME: hack to comply with ANSI rules */
	int t=tell();
	fseek(file, 0, SEEK_END);	/* zero is allowed ;-) */
	int r=ftell(file);
	fseek(file, t, SEEK_SET);
	return r;
}

dword ht_file_stream::read(void *buf, dword size)
{
	dword r=fread(buf, 1, size, file);
	offset+=r;
	return r;
}

int ht_file_stream::seek(STREAMOFS _offset)
{
	if (_offset==offset) return 0;
	int r=fseek(file, _offset, SEEK_SET);
	if (!r) offset=_offset;
	return r;
}

int ht_file_stream::set_access_mode(dword _access_mode)
{
	dword orig_access_mode=access_mode;
	int r=set_access_mode_internal(_access_mode);
	if (!r) set_access_mode_internal(orig_access_mode);
	return r;
}

int ht_file_stream::set_access_mode_internal(dword _access_mode)
{
RETRY:
	if (access_mode==_access_mode) return 1;
	char *mode=0;
	if (_access_mode & SAM_CREATE) {
		if (_access_mode & SAM_WRITE) mode="wb";
		if (_access_mode & SAM_READ) mode="wb+";
	} else {
		if (_access_mode & SAM_READ) mode="rb";
		if (_access_mode & SAM_WRITE) mode="rb+";
	}
	if (!mode) {
		set_error(EINVAL | STERR_SYSTEM); // FIXME: stream error msgs
		return 0;
	}
	if (file) fclose(file);
	access_mode=SAM_UNDEFINED;
	file=fopen(name, mode);
	if (!file) {
		set_error(errno | STERR_SYSTEM);
		if ((stream_error_proc) && (stream_error_proc(this)==SERR_RETRY))
			goto RETRY;
		return 0;
	}
	ht_stream::set_access_mode(_access_mode);
	return 1;
}

void ht_file_stream::pstat(ht_pstat *s)
{
	struct stat st;
	fstat(fileno(file), &st);
#ifdef DJGPP
	s->caps=pstat_mtime|pstat_mode_usr|pstat_mode_w|pstat_size/*|pstat_cluster*/;
	s->mtime=st.st_mtime;
	s->mode=st.st_mode;
	s->size=st.st_size;
	s->fsid=st.st_ino;
#else
	s->caps=pstat_ctime|pstat_mtime|pstat_atime|pstat_uid|pstat_gid|pstat_mode_all|pstat_size|pstat_inode;
	s->ctime=st.st_ctime;
	s->mtime=st.st_mtime;
	s->atime=st.st_atime;
	s->gid=st.st_uid;
	s->uid=st.st_gid;
	s->mode=st.st_mode;
	s->size=st.st_size;
	s->fsid=st.st_ino;
#endif
}

STREAMOFS ht_file_stream::tell()
{
	return offset;
}

dword ht_file_stream::write(void *buf, dword size)
{
	dword r=fwrite(buf, 1, size, file);
	offset+=r;
	return r;
}

/*
 *	CLASS ht_mem_stream
 */

#define HTMEMFILE_INITIAL_SIZE 1024
#define HTMEMFILE_GROW_FACTOR 1.5

void ht_mem_stream::init()
{
	ht_stream::init();
	ofs=0;
	bufsize=HTMEMFILE_INITIAL_SIZE;
	buf=(byte*)malloc(bufsize);
	memset(buf, 0, bufsize);
	pos=0;
	dsize=0;
}

void ht_mem_stream::init(STREAMOFS _ofs, dword size)
{
	ht_stream::init();
	ofs=_ofs;
	buf=(byte*)malloc(size);
	bufsize=size;
	memset(buf, 0, size);
	pos=0;
	dsize=0;
}

void ht_mem_stream::done()
{
	delete buf;
	ht_stream::done();
}

void ht_mem_stream::extendbuf()
{
	dword oldbufsize=bufsize;
//	bufsize+=1024;
	bufsize=(dword)(bufsize*HTMEMFILE_GROW_FACTOR);
	byte *t=(byte*)malloc(bufsize);
	memset(t, 0, bufsize);
	memmove(t, buf, oldbufsize);
	delete buf;
	buf=t;
}

dword ht_mem_stream::get_size()
{
	return dsize;
}

char *ht_mem_stream::get_name()
{
	return "memstream";
}

dword ht_mem_stream::read(void *buf, dword size)
{
	if (pos+size>bufsize) {
		memset(buf, 0, size);
		if (bufsize-pos+1>0) memmove(buf, this->buf+pos, bufsize-pos+1);
		size=bufsize-pos;
	} else {
		memmove(buf, (char*)this->buf+pos, size);
	}
	pos+=size;
	return size;
}

int ht_mem_stream::seek(STREAMOFS _offset)
{
	if (_offset<ofs)
		return 0;
	pos=_offset-ofs;
	return 1;
}

STREAMOFS ht_mem_stream::tell()
{
	return pos+ofs;
}

dword ht_mem_stream::write(void *_buf, dword size)
{
	while (pos+size>=bufsize) extendbuf();
	memmove(((byte*)buf)+pos, _buf, size);
	pos+=size;
	if (pos>dsize) dsize=pos;
	return size;
}

/*
 *	CLASS ht_cache_page
 */

ht_cache_page::ht_cache_page(STREAMOFS _offset, dword _size, dword _alloc_index)
{
	data=(byte*)malloc(_size);
	offset=_offset;
	size=_size;
	alloc_index=_alloc_index;
	read_count=0;
	write_count=0;
}

ht_cache_page::~ht_cache_page()
{
	free(data);
}

int ht_cache_page::read(STREAMOFS offset, byte *buf, dword len)
{
	if (offset<size) {
		dword c=(offset+len>size) ? size-offset : len;
		memmove(buf, data+offset, c);
//		read_count+=c;
		return c;
	}
	return 0;
}

int ht_cache_page::readbyte(byte *b, STREAMOFS offset)
{
	if (offset<size) {
		*b=data[offset];
		read_count++;
		return 1;
	}
	return 0;
}

int ht_cache_page::writebyte(byte b, STREAMOFS offset)
{
	if (offset<size) {
		data[offset]=b;
		write_count++;
		return 1;
	}
	return 0;
}

/*
 *	CLASS ht_cached_stream
 */

#define STD_CACHE_PAGE_SIZE 512	/* MUST BE 2^n */

void ht_cached_stream::init(ht_stream *stream, int own_stream, dword _maxcachesize, int _flags)
{
	ht_layer_stream::init(stream, own_stream);
	cache=new ht_dtree();
	cache->init(compare_keys_uint);
	maxcachesize=_maxcachesize;
	flags=_flags;
	if (!maxcachesize) flags&=~(CACHED_STREAM_CACHE_READS | CACHED_STREAM_CACHE_WRITES);
	if (flags & CACHED_STREAM_CACHE_READS) mode|=SM_READ_CACHED;
	alloc_index=0;
	offset=0;
	if (!stream->get_error()) size=stream->get_size();
	read_pages=0;
	write_pages=0;
	page_granularity=STD_CACHE_PAGE_SIZE;
	page_mask=~(page_granularity-1);
}

void ht_cached_stream::init(ht_stream *stream, int own_stream, dword _maxcachesize, int _flags, dword _page_granularity)
{
	ht_layer_stream::init(stream, own_stream);
	cache=new ht_dtree();
	cache->init(compare_keys_uint);
	maxcachesize=_maxcachesize;
	flags=_flags;
	if (!maxcachesize) flags&=~(CACHED_STREAM_CACHE_READS | CACHED_STREAM_CACHE_WRITES);
	if (flags & CACHED_STREAM_CACHE_READS) mode|=SM_READ_CACHED;
	alloc_index=0;
	offset=0;
	if (!stream->get_error()) size=stream->get_size();
	read_pages=0;
	write_pages=0;
	page_granularity=_page_granularity;
	page_mask=~(page_granularity-1);
}

void ht_cached_stream::done()
{
	cache->done();
	delete cache;
	ht_layer_stream::done();
}

ht_cache_page *ht_cached_stream::cache_page_create(STREAMOFS offset)
{
	offset&=page_mask;
	dword size=stream->get_size();
	if (offset>size) return 0;
	size-=offset;
	if (size>page_granularity) size=page_granularity;
	ht_cache_page *c=new ht_cache_page(offset, size, alloc_index++);
	stream->seek(offset);
	stream->read(c->data, size);
	cache->insert(new ht_data_uint(offset), c);
	return c;
}

void ht_cached_stream::cache_page_destroy(ht_cache_page *c)
{
	ht_data_uint o;
	o.value=offset;
	cache->del(&o);
}

ht_cache_page *ht_cached_stream::cache_page_find(STREAMOFS offset)
{
	ht_data_uint o;
	o.value=offset & page_mask;
	return (ht_cache_page*)cache->get(&o);
}

void ht_cached_stream::cache_page_flush(ht_cache_page *c)
{
	if ((c->write_count) && (flags & CACHED_STREAM_CACHE_WRITES)) {
		stream->seek(c->offset);
		stream->write(c->data, c->size);
	}
//	c->read_count=0;
	c->write_count=0;
}

void ht_cached_stream::flush()
{
	ht_data *key, *data;
	bool b=cache->enum_first(&key, &data);
	while (b) {
		ht_cache_page *c=(ht_cache_page*)data;
		cache_page_flush(c);
		b=cache->enum_next(&key, &data);
	}
}

void ht_cached_stream::flush_and_free_least_used_page()
{
	ht_cache_page *lup=0;
	dword lup_access_count=0x7fffffff;
	dword lup_alloc_index=0x7fffffff;
	ht_data *key, *data;
	bool b=cache->enum_first(&key, &data);

	ht_cache_page *c;
	while (b) {
		c=(ht_cache_page*)data;
		if (((flags & CACHED_STREAM_ONLY_EXPLICIT_FLUSHES) && (!c->write_count)) ||
		!(flags & CACHED_STREAM_ONLY_EXPLICIT_FLUSHES)) {
			dword access_count=c->write_count+c->read_count;
			if (access_count<lup_access_count) {
				lup_access_count=access_count;
				lup_alloc_index=c->alloc_index;
				lup=c;
			} else if ((access_count==lup_access_count) &&
			  (c->alloc_index<lup_alloc_index)) {
				lup_alloc_index=c->alloc_index;
				lup=c;
			}
		}
		b=cache->enum_next(&key, &data);
	}
	if (lup) {
		cache_page_flush(lup);
		cache_page_destroy(lup);
	}
}

void ht_cached_stream::free_read_pages()
{
/*restart:
	ht_node *n=cache->enum_first();
	ht_cache_page *c;
	while (n) {
		c=(ht_cache_page*)n->data;
		if (!c->write_count) {
			cache->del(n->key);
			goto restart;
		}
		n=cache->enum_next();
	}
	read_pages=0;*/
}

dword ht_cached_stream::get_size()
{
	return size;
}

void ht_cached_stream::invalidate()
{
	cache->done();
	delete cache;
	cache=new ht_dtree();
	cache->init(compare_keys_uint);
}

dword ht_cached_stream::read(void *buf, dword size)
{
	int c=0;
	if (flags & CACHED_STREAM_CACHE_READS) {
		if (size>4) {
			ht_cache_page *p;
			STREAMOFS o=offset;
			while (size) {
				p=cache_page_find(o);
				if (!p) p=read_miss(o);
				if (!p) return c;
				dword lc=page_granularity - (o & (~page_mask));
				if (lc>size) lc=size;
				dword k=p->read(o & (~page_mask), (byte*)buf, lc);
				if (!k) break;
				size-=k;
				o+=k;
				c+=k;
				(byte*)buf+=k;
			}
			offset=o;
			return c;
		} else {
			for (dword i=0; i<size; i++) {
				if (!readbyte(offset+i, ((byte*)buf)+i)) break;
				c++;
			}
			offset+=c;
			return c;
		}
	}
	stream->seek(offset);
	c=stream->read(buf, size);
	offset+=c;
	return c;
}

int ht_cached_stream::readbyte(STREAMOFS offset, byte *b)
{
	if (flags & CACHED_STREAM_CACHE_READS) {
		ht_cache_page *c=cache_page_find(offset);
		if (!c) c=read_miss(offset);
		return c->readbyte(b, offset & (~page_mask));
	}
	stream->seek(offset);
	return stream->read(b, 1);
}

ht_cache_page *ht_cached_stream::read_miss(STREAMOFS offset)
{
	if (flags & CACHED_STREAM_ONLY_EXPLICIT_FLUSHES) {
		if (read_pages*(page_granularity+sizeof (ht_cache_page))>maxcachesize) flush_and_free_least_used_page();
	} else {
		if ((read_pages+write_pages)*(page_granularity+sizeof (ht_cache_page))>maxcachesize) flush_and_free_least_used_page();
	}
	ht_cache_page *c=cache_page_create(offset);
	if (c) read_pages++;
	return c;
}

int ht_cached_stream::seek(STREAMOFS _offset)
{
	if (_offset<size) {
		offset=_offset;
		return 1;
	}
	return 0;
}

void ht_cached_stream::set_mode(dword _mode)
{
	if ((_mode & SM_READ_CACHED) && !(flags & CACHED_STREAM_CACHE_READS)) {
/* enable read caching */
		mode|=SM_READ_CACHED;
		flags|=CACHED_STREAM_CACHE_READS;
	} else if (!(_mode & SM_READ_CACHED) && (flags & CACHED_STREAM_CACHE_READS)) {
/* disable read caching */
		mode&=~SM_READ_CACHED;
		flags&=~CACHED_STREAM_CACHE_READS;
	}
}

STREAMOFS ht_cached_stream::tell()
{
	return offset;
}

dword ht_cached_stream::write(void *buf, dword size)
{
	int c=0;
	for (dword i=0; i<size; i++) {
		if (!writebyte(offset+i, *(((byte*)buf)+i))) break;
		c++;
	}
	offset+=c;
	return c;
}

int ht_cached_stream::writebyte(STREAMOFS offset, byte b)
{
	if (flags & CACHED_STREAM_CACHE_WRITES) {
		ht_cache_page *c=cache_page_find(offset);
		if (!c) {
			if (flags & CACHED_STREAM_ONLY_EXPLICIT_FLUSHES) {
//          		if (read_pages*(page_granularity+sizeof (ht_cache_page))>maxcachesize) flush_and_free_least_used_page();
			} else {
				if ((read_pages+write_pages)*(page_granularity+sizeof (ht_cache_page))>maxcachesize) flush_and_free_least_used_page();
			}
			c=cache_page_create(offset);
			if (!c) return 0;
			write_pages++;
		} else if (!c->write_count) {
			read_pages--;
			write_pages++;
		}
		return c->writebyte(b, offset & (~page_mask));
	}
	stream->seek(offset);
	return stream->write(&b, 1);
}

/*
 *	CLASS ht_cache_page_pbd
 */

ht_cache_page_pbd::ht_cache_page_pbd(STREAMOFS offset, dword size, dword alloc_index, ht_stream *_stream)
: ht_cache_page(offset, size, alloc_index)
{
	writedirty=(byte*)malloc(size);
	stream=_stream;
	cleanup();
}

ht_cache_page_pbd::~ht_cache_page_pbd()
{
	delete writedirty;
}

int ht_cache_page_pbd::isdirty(STREAMOFS offset, int range)
{
	if (range & ISDIRTY_SINGLEBIT) {
		dword m=1<<(range & ~ISDIRTY_SINGLEBIT);
		return ((writedirty[offset] & m) != (data[offset] & m));
	} else {
		for (int i=0; i<range; i++) {
			if (offset+i>=size) break;
			if (writedirty[offset+i]!=data[offset+i]) return 1;
		}
	}
	return 0;
}

void ht_cache_page_pbd::cleanup()
{
	STREAMOFS o=stream->tell();
	stream->seek(offset);
	stream->read(writedirty, size);
	stream->seek(o);
}

/*
 *	CLASS ht_cached_stream_pbd
 */

void ht_cached_stream_pbd::init(ht_stream *stream, int own_stream, dword maxcachesize, int flags)
{
	ht_cached_stream::init(stream, own_stream, maxcachesize, flags);
}

void ht_cached_stream_pbd::init(ht_stream *stream, int own_stream, dword maxcachesize, int flags, dword page_granularity)
{
	ht_cached_stream::init(stream, own_stream, maxcachesize, flags, page_granularity);
}

ht_cache_page *ht_cached_stream_pbd::cache_page_create(STREAMOFS offset)
{
	offset&=page_mask;
	dword size=stream->get_size()-offset;
	if (size>page_granularity) size=page_granularity;
	ht_cache_page_pbd *c=new ht_cache_page_pbd(offset, size, alloc_index++, stream);
	stream->seek(offset);
	stream->read(c->data, size);
	cache->insert(new ht_data_uint(offset), c);
	return c;
}

void ht_cached_stream_pbd::cache_page_flush(ht_cache_page *c)
{
	ht_cached_stream::cache_page_flush(c);
	((ht_cache_page_pbd*)c)->cleanup();
}

int ht_cached_stream_pbd::isdirty(STREAMOFS offset, int range)
{
	if (range & ISDIRTY_SINGLEBIT) {
		ht_cache_page *c;
		c=cache_page_find(offset);
		if (c) {
			if (((ht_cache_page_pbd*)c)->isdirty(offset&(~page_mask), range)) return 1;
		}
	} else {
		ht_cache_page *c;
		int pr;
		while (range>0) {
			c=cache_page_find(offset);
			if (c) {
				pr=range;
				if ((offset&(~page_mask))+pr>c->size) pr=c->size-(offset&(~page_mask));
				if (((ht_cache_page_pbd*)c)->isdirty(offset&(~page_mask), pr)) return 1;
				offset+=pr;
				range-=pr;
			} else {
				offset+=page_granularity;
				range-=page_granularity-(offset&(~page_mask));
				offset&=page_mask;
			}
		}
	}
	return 0;
}

/*
 *	string stream functions
 */

// FIXME: more dynamical solution appreciated
#define REASONABLE_STRING_LIMIT	256
 
char *getstrz(ht_stream *stream)
{
/* get string size */
	char buf[64];
	int s, z=0, o=stream->tell();
	int found=0;
	while (!found) {
		if (!(s=stream->read(buf, 64))) break;
		for (int i=0; i<s; i++) {
			z++;
			if (!buf[i]) {
				found=1;
				break;
			}
		}
	}
/* read string */
	if (z>REASONABLE_STRING_LIMIT) z=REASONABLE_STRING_LIMIT;
	stream->seek(o);
	char *str=(char *)malloc(z);
	stream->read(str, z);
	str[z-1]=0;
	return str;
}

void putstrz(ht_stream *stream, char *str)
{
	stream->write(str, strlen(str)+1);
}

char *getstrp(ht_stream *stream)
{
	unsigned char l;
	stream->read(&l, 1);
	char *str=(char*)malloc(l+1);
	stream->read(str, l);
	*(str+l)=0;
	return str;
}

void putstrp(ht_stream *stream, char *str)
{
	unsigned char l=strlen(str);
	stream->write(&l, 1);
	stream->write(str, l);
}

char *getstrw(ht_stream *stream)
{
	short t;
	int l=0;
	stream->read(&l, 2);
	char	*a=(char*)malloc(l+1);
	for (int i=0; i<l; i++) {
		stream->read(&t, 2);
		*(a+i)=(char)t;
	}
	*(a+l)=0;
	return a;
}

void putstrw(ht_stream *stream, char *str)
{
	/* FIXME: someone implement me ? */
}

/*
 *   file system functions
 */

#if defined(WIN32) || defined(__WIN32__)
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>
#elif defined(DJGPP)
#include <sys/stat.h>
#else
#include <stdlib.h>
#include <unistd.h>
#endif

int canonicalize(char *in_file, char *out_file)
{
#if defined(WIN32) || defined(__WIN32__)
	char *dunno;
	return GetFullPathName(in_file, 255, out_file, &dunno) > 0;
#elif defined(DJGPP)
	_fixpath(in_file, out_file);
	return 1;
#else
	return (realpath(in_file, out_file)==out_file);
#endif
#if 0
	strcpy(out_file, in_file);
	return 1;
#endif
}

